#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "misc.h"
#include "lxmisc.h"
#include <lx_map.h>

#include "naw_log.h"


/***************************************************************************
  default log pointer -> send as debugging message
 ***************************************************************************/

static int naw_logptr_default(level, s)
    int level;
    lx_s *s;
{
    switch (level) {
        case NAW_DEBUG:  level = 1; break;
        case NAW_DEBUG2: level = 2; break;
        case NAW_DEBUG3: level = 3; break;
        default:         level = 0; break;
    }
    if (lx_check0(s)) return -1;
    log_debug(level, s->s, 0,0,0,0);
    return 0;
}


/* XXX: not thread safe.  But we shouldn't perform simultaneous logging
 * anyway, so locking can be done in the logging function.
 */
static lx_s logs = {0};

int (*naw_logptr)(int, lx_s *) = naw_logptr_default;

naw_logptr_t naw_log_register(naw_logptr_t new) {
    naw_logptr_t old = naw_logptr;
    naw_logptr = new;
    return old;
}

static inline int naw_logptr_send(int level, lx_s *s) {
    if ((naw_debug_level+NAW_DEBUG) > level)
        return naw_logptr(level, s);
    return 0;
}

/***************************************************************************
  mapped logging function.  Argument format:
    LXM_q("quoted string"):  direct quoted string (NOT pointer to data)
    LXM_qm("quoted memory"): direct quoted memory (NOT pointer to data)
    LXM_mb(memptr, len):     memory pointer, and the length to use
    LXM_l(number):           long decimal
    LXM_d(number):           int decimal
    LXM_lu(number):          unsigned long decimal
    LXM_du(number):          unsigned int  decimal
    LXM_lxs(s):              pointer to lx_s (lx_s *)
    LXM_END:                 end of argument list (this MUST be included)

  See lx_map.h of the lx_lib package for more details.

  NAW_LOG() is a wrapper around naw_log() that includes LXM_END at the end
  of the argument list, so you need not do so yourself.

  Note: The correct way of passing a pointer to memory is LXM_mb(p, len).
  If you must have something along the lines of '%s' functionality, you can
  define LXM_s(p) to LXM_mb(p, strlen(p)) yourself.

  Generally speaking, this method is faster and safer than the use of format
  strings.  You may or may not consider it uglier.
 ***************************************************************************/

int naw_log(int level, ...) {
    va_list ap;
    int r, sw;

    logs.len=0;
    va_start(ap, level);

    for (;;) {
        sw = lx_map(lx_striadd, &logs, &ap, &r);
        if (!sw) continue;

        va_end(ap);
        if (sw < 0) return -1;
        break;
    }

    return naw_logptr_send(level, &logs);
}

/***************************************************************************
  Format string -based logger function.

  Only basic formatting functionality is supported (prefixes are not).
  Currently supported format strings (see printf(3) for more details):
    %d
    %ld
    %lu
    %x
    %p
    %s
    %c
    %%
    %m (nonstandard): log message string (see syslog(3))
 ***************************************************************************/

/* XXX: separate out so va_end() can be used properly */
int naw_logfmt(int level, char *fmt, ...){
    va_list ap;
    char *seg = fmt;
    register char c;

    logs.len=0;
    va_start(ap, fmt);

#define NAW_ADDLX if (lx_striadd(&logs, seg, fmt-seg)) return -1

    for (;;) {
        c = *fmt;
        if (!c) {
            NAW_ADDLX;
            va_end(ap);
            return naw_logptr_send(level, &logs);
        } else if (c == '%') {
            NAW_ADDLX;
            c = *++fmt;
            if (c == '%') {
                if (lx_cadd(&logs, '%'))
                    return -1;
            } else if (c == 's') {
                if (lx_stradd(&logs, va_arg(ap, char *)))
                    return -1;
            } else if (c == 'd') {
                if (lx_straddlong(&logs, va_arg(ap, int), 10))
                    return -1;
            } else if (c == 'u') {
                if (lx_straddulong(&logs, va_arg(ap, unsigned int), 10))
                    return -1;
            } else if (c == 'l') {
                switch (*++fmt) {
                    case 'u':
                        if (lx_straddulong(&logs,
                                    va_arg(ap, unsigned long), 10))
                            return -1;
                        break;
                    case 'd':
                        if (lx_straddlong(&logs, va_arg(ap, long), 10))
                            return -1;
                        break;
                    default:
                        errno = EINVAL;
                        return -1;
                }
            } else if (c == 'x') {
                if (lx_straddulong(&logs, va_arg(ap, unsigned long), 16))
                    return -1;
            } else if (c == 'p') {
                if (lx_striadd(&logs, "0x", 2) ||
                    lx_straddulong(&logs, (unsigned long)va_arg(ap,void*), 16))
                    return -1;
            } else if (c == 'm') {
                if (lx_stradd(&logs, strerror(errno)))
                    return -1;
            } else if (c == 'c') {
                if (lx_cadd(&logs, (char)va_arg(ap, int)))
                    return -1;
            }
            /* XXX: maybe float? */
            seg = fmt+1;
        }

    fmt++;
    }
}

#if 0
//#define AMT 10000000
#define AMT 5
#define HOST "abooga.abooga.com"
int main(void) {
    int n;

    for (n = 0; n < AMT; n++) {
        NAW_LOG(3,
                LXM_q("Accepting connection from ("),
                LXM_q(HOST),
                LXM_q(") on port "),
                LXM_l(543));

        naw_logfmt(3, "Accepting connection from (%s) on port %ld",
                HOST, 543);
    }

    return 0;
}
#endif

