#include <stdio.h>
#include <cryptlib.h>
#include "misc.h"
#include "naw.h"

static int status = !CRYPT_OK;


static int CRYPT_FAIL(session, keyset)
    CRYPT_SESSION *session;
    CRYPT_KEYSET *keyset;
{
    if (session) cryptDestroySession(*session);
    if (keyset)  cryptKeysetClose (*keyset);
    return NAW_ERR;
}

/***************************************************************************
  naw_cert_auth_accept

  takes the place of both 'accept' and 'reject' actions.  Only returns YES,
  so 'deny' actions will result in rejection, and 'accept' actions will
  result in access.
 ***************************************************************************/

static int naw_cert_auth_accept (handle, data)
    struct naw_handle *handle;
    struct naw_data *data;
{
    char *v;
    int vlen;
    int ssl_type, r;

    if (handle->type == INIT_CLI) ssl_type = CRYPT_SESSION_SSL;
    else                          ssl_type = CRYPT_SESSION_SSL_SERVER;

    CRYPT_SESSION cryptSession;

    /* XXX: multiple inits are okay, too */
    if (status != CRYPT_OK) {
        log_debug (1,"cryptlib initialization failed; cryptlib is unusable",0,0,0,0);
        return CRYPT_FAIL(0, 0);
    }

    if (cryptCreateSession(&cryptSession, CRYPT_UNUSED, ssl_type) != CRYPT_OK) {
        log_debug(1,"cryptCreateSession() failed",0,0,0,0);
        return CRYPT_FAIL(0, 0);
    }

    if (cryptSetAttribute(cryptSession,
                CRYPT_SESSINFO_NETWORKSOCKET, handle->sd) != CRYPT_OK) {
        log_debug(1,"cryptSetAttribute(CRYPT_SESSINFO_NETWORKSOCKET) failed",0,0,0,0);
        return CRYPT_FAIL(&cryptSession, 0);
    }

    if (!naw_data_get(data, "SSLSharedKey", &v, &vlen)) {
        /* XXX: expand... allow for file input? */
        log_debug (3,"allowing shared key: ",v,0,0,0);
        cryptSetAttributeString(cryptSession,
                CRYPT_SESSINFO_USERNAME, "default", 7);
        cryptSetAttributeString(cryptSession,
                CRYPT_SESSINFO_PASSWORD, v, vlen);
    }

    if (!naw_data_get(data, "SSLPrivateKey", &v, &vlen)) {
        log_debug (3,"allowing public key: ",v,0,0,0);
        /* XXX: implement */
    }

    if (handle->type == INIT_SRV && !naw_data_get(data, "SSLCertFile",&v, &vlen)) {
        CRYPT_KEYSET cks;

        log_debug (3,"gathering certificate file: ",v,0,0,0);
        r = cryptKeysetOpen(&cks, CRYPT_UNUSED, CRYPT_KEYSET_FILE, v,
                CRYPT_KEYOPT_READONLY);

        if (r != CRYPT_OK) {
            log_debug(1,"error opening keyset file ",v,0,0,0);
            fprintf(stderr, "returned: %d\n", r); fflush(stderr);
            return CRYPT_FAIL(&cryptSession, 0);
        }

        r = cryptSetAttribute(cryptSession, CRYPT_SESSINFO_CMP_PRIVKEYSET, cks);

        if (r != CRYPT_OK) {
            log_debug(1,"error reading keyset file ",v,0,0,0);
            fprintf(stderr, "returned: %d\n", r); fflush(stderr);
            return CRYPT_FAIL(&cryptSession, &cks);
        }

        cryptKeysetClose(cks);
    } else if (handle->type != INIT_CLI) {
        log_debug(1,"certauth: missing SSLCertFile",0,0,0,0);
        return NAW_ERR;
    }

    r = cryptSetAttribute(cryptSession, CRYPT_SESSINFO_ACTIVE, 1);

    cryptDestroySession(cryptSession);

    switch (r) {
        case CRYPT_OK:
            log_debug (2, "session authentication succeeded",0,0,0,0);
            return NAW_YES;
        //default:
        /* XXX: check for no-auth error code */
    }
    log_debug (2,"session authentication failed",0,0,0,0);
    return NAW_NO;
}

static void naw_cert_auth_init (void)
{
    status = cryptInit();
    cryptAddRandom(0, CRYPT_RANDOM_FASTPOLL);
}

static void naw_cert_auth_fini (void)
{
    if (cryptEnd() != CRYPT_OK)
        log_debug(1,"cryptlib shutdown failed",0,0,0,0);;
}

struct naw_modfuncset naw_set = {
    naw_cert_auth_accept,  /* accept  */
    naw_cert_auth_accept,  /* connect */
    naw_cert_auth_init,    /* init    */
    naw_cert_auth_fini,    /* fini    */
};

