#include <openssl/err.h>
#include <openssl/ssl.h>
#include <openssl/rand.h>
#include <stdio.h>
#include <string.h>
#include "misc.h"
#include "naw.h"
#include "naw_log.h"

/* XXX we need:
     - a CRL implementation
*/

struct naw_x509_alloc {
    SSL_CTX *ctx;
    SSL *ssl;
};

static inline int freeret (a, r)
    struct naw_x509_alloc *a;
    int r;
{
    int errcode;

    /* XXX multithreading: check implications for ERR_get_error() */
    while ((errcode = ERR_get_error()))
        naw_logfmt(NAW_WARNING, "x509: %s", ERR_error_string(errcode, 0));

    if (a->ssl) { SSL_free(a->ssl); a->ssl = 0; }
    if (a->ctx) { SSL_CTX_free(a->ctx); a->ctx = 0; }
    return r;
    /* XXX: deal with error stack, too */
}

static inline int naw_x509_setup(a)
    struct naw_x509_alloc *a;
{
    memset(a, 0, sizeof(struct naw_x509_alloc));

    a->ctx = SSL_CTX_new(TLS_method());
    if (!a->ctx) {
        NAW_LOG(NAW_WARNING, LXM_q("x509: creation of new ctx failed"));
        return freeret(a, NAW_ERR);
    }

    a->ssl = SSL_new(a->ctx);
    if (!a->ssl) {
        NAW_LOG(NAW_WARNING, LXM_q("x509: creation of new ssl failed"));
        return freeret(a, NAW_ERR);
    }

    return NAW_OK;
}

/* TLS adds some application messages after the handshake completes.  They
 * aren't useful to us, so here we just slurp them up. Any errors in the
 * process (e.g. if the client cert verification failed) are caught in
 * freeret(). */
static inline void slurp_buf(SSL *a) {
    char buf[2048];
    int r;
    for (;;) {
        r = SSL_read(a, buf, sizeof(buf));
        if (r <= 0)
            return;  /* XXX: check error to see if this is really the end */
        NAW_LOG(NAW_DEBUG, LXM_q("read SSL data: "), LXM_mb(buf, r));
    }
}

/***************************************************************************
  naw_test_module_accept

  takes the place of both 'accept' and 'reject' actions.  Only returns YES,
  so 'deny' actions will result in rejection, and 'accept' actions will
  result in access.
 ***************************************************************************/

static int naw_x509_accept (handle, data)
    struct naw_handle *handle;
    struct naw_data *data;
{
    struct naw_x509_alloc a;
    char *v;
    unsigned int vlen;
    char *keyfile, *certfile, *cafile, *crlfile;
    void (*ssl_cmd)(SSL *ssl);

    if (handle->type == INIT_SRV) {
        keyfile  = "x509ServerKeyFile";
        certfile = "x509ServerCertFile";
        cafile   = "x509ServerCAFile";
        crlfile  = "x509ServerCRLFile";
        ssl_cmd  = SSL_set_accept_state;
    } else {
        keyfile  = "x509ClientKeyFile";
        certfile = "x509ClientCertFile";
        cafile   = "x509ClientCAFile";
        crlfile  = "x509ClientCRLFile";
        ssl_cmd  = SSL_set_connect_state;
    }
#define common_keyfile  "x509KeyFile"
#define common_certfile "x509CertFile"
#define common_cafile   "x509CAFile"
#define common_crlfile  "x509CRLFile"

    memset(&a, 0, sizeof(struct naw_x509_alloc));

    a.ctx = SSL_CTX_new(TLS_method());
    if (!a.ctx) {
        NAW_LOG(NAW_WARNING, LXM_q("x509: creation of new ctx failed"));
        goto ERR_RETURN;
    }

    if (!naw_data_get(data, keyfile, &v, &vlen) ||
        !naw_data_get(data, common_keyfile, &v, &vlen)) {
        NAW_LOG(NAW_DEBUG3, LXM_q("x509: using key file: "),
                LXM_mb(v,vlen));
        if (!SSL_CTX_use_PrivateKey_file(a.ctx, v, SSL_FILETYPE_PEM)) {
            NAW_LOG(NAW_NOTICE,
                    LXM_q("x509: error reading private key file "),
                    LXM_mb(v,vlen));
            goto ERR_RETURN;
        }
    }

    if (!naw_data_get(data, certfile, &v, &vlen) ||
        !naw_data_get(data, common_certfile, &v, &vlen)) {
        NAW_LOG(NAW_DEBUG3, LXM_q("x509: using cert file: "),
                LXM_mb(v,vlen));
        if (!SSL_CTX_use_certificate_file(a.ctx, v, SSL_FILETYPE_PEM)) {
            NAW_LOG(NAW_NOTICE,
                    LXM_q("x509: error reading certificate file "),
                    LXM_mb(v,vlen));
            goto ERR_RETURN;
        }
    }

    if (!naw_data_get(data, cafile, &v, &vlen) ||
        !naw_data_get(data, common_cafile, &v, &vlen)) {
        NAW_LOG(NAW_DEBUG3, LXM_q("x509: using CA file: "),
                LXM_mb(v,vlen));
        if (!SSL_CTX_load_verify_locations(a.ctx, v, 0)) {
            NAW_LOG(NAW_NOTICE, LXM_q("x509: error reading CA file "),
                    LXM_mb(v,vlen));
            goto ERR_RETURN;
        }
    }

    if (!naw_data_get(data, crlfile, &v, &vlen) ||
        !naw_data_get(data, common_crlfile, &v, &vlen)) {

        X509_STORE *store = SSL_CTX_get_cert_store(a.ctx);
        X509_VERIFY_PARAM *param;
        X509_CRL *crl;
        FILE *fp;

        NAW_LOG(NAW_DEBUG3, LXM_q("x509: using CRL file: "),
                LXM_mb(v,vlen));
        if (!(fp = fopen(v, "r"))) {
            NAW_LOG(NAW_NOTICE, LXM_q("x509: error opening CRL file "),
                    LXM_mb(v,vlen));
            goto ERR_RETURN;
        }

        crl = PEM_read_X509_CRL(fp, 0, 0, 0);
        fclose(fp);
        if (!crl) {
            NAW_LOG(NAW_NOTICE, LXM_q("x509: error reading CRL file "),
                    LXM_mb(v,vlen), LXM_q(" (NOTE: must be PEM format)"));
            goto ERR_RETURN;
        }
        if (X509_STORE_add_crl(store, crl) != 1) {
            NAW_LOG(NAW_NOTICE, LXM_q("x509: BUG: error adding CRL to store"));
            X509_CRL_free(crl);
            goto ERR_RETURN;
        }
        X509_CRL_free(crl);

        param = X509_VERIFY_PARAM_new();
        if (!param) {
            NAW_LOG(NAW_WARNING, LXM_q("x509: generic param creation failed: OOM?"));
            goto ERR_RETURN;
        }
        /* XXX: shouldn't need  ...CRL_CHECK_ALL here, but see NOTES */
        if (
                X509_VERIFY_PARAM_set_flags(param, X509_V_FLAG_CRL_CHECK) != 1
                || X509_STORE_set1_param(store, param) != 1 ) {
            NAW_LOG(NAW_WARNING, LXM_q("x509: BUG: failed to set CRL params"));
            X509_VERIFY_PARAM_free(param);
            goto ERR_RETURN;
        }
        X509_VERIFY_PARAM_free(param);
    }


    a.ssl = SSL_new(a.ctx);
    if (!a.ssl) {
        NAW_LOG(NAW_WARNING, LXM_q("x509: creation of new ssl failed"));
        goto ERR_RETURN;
    }


    SSL_set_fd(a.ssl, handle->sd);

    SSL_set_verify(a.ssl, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT, 0);

    /* set connect or accept state before handshake */
    ssl_cmd(a.ssl);

    if (SSL_do_handshake(a.ssl) != 1) {
        /* XXX: use SSL_get_verify_result() to log failure reason */
        NAW_LOG(NAW_INFO, LXM_q("x509: ssl negotiation failed"));
        goto ERR_RETURN;
    }
    NAW_LOG(NAW_INFO, LXM_q("x509: ssl negotiation succeeded"));

    if (handle->type == INIT_SRV) {
        X509 *x509_client = SSL_get_peer_certificate(a.ssl);
        X509_NAME *x509_sn;
        char sn[80];

        if (!x509_client) {
            NAW_LOG(NAW_WARNING, LXM_q("x509: BUG: ssl negotiation succeeded but no client cert"));
            goto ERR_RETURN;
        }

        x509_sn = X509_get_subject_name(x509_client);
        /* XXX: X509_NAME_oneline() is "discouraged". Alternative? */
        if (!X509_NAME_oneline(x509_sn, sn, sizeof(sn)))
            goto ERR_RETURN;

        naw_logfmt(NAW_NOTICE, "x509: access granted for [%s]", sn);

        X509_free(x509_client);
    }

    SSL_shutdown(a.ssl);
    slurp_buf(a.ssl);
    return freeret(&a, NAW_YES);

ERR_RETURN:
    return freeret(&a, NAW_ERR);
}

static int naw_x509_init (data)
    struct naw_data *data;
{
    NAW_LOG (NAW_DEBUG2, LXM_q("x509: initializing"));
    SSL_load_error_strings();
    SSL_library_init();
    if (!RAND_status()) {
        NAW_LOG(NAW_NOTICE, LXM_q("x509: not enough random data seeded"));
        /* XXX: seed prng */
    }
    return NAW_OK;
}

static void naw_x509_fini (void)
{
    NAW_LOG (NAW_DEBUG2, LXM_q("x509: unloading"));
    ERR_free_strings();
    /* XXX: RAND_cleanup(); ? */
}

struct naw_modfuncset naw_set = {
    naw_x509_accept,  /* accept  */
    naw_x509_accept,  /* connect */
    naw_x509_init,    /* init    */
    naw_x509_fini,    /* fini    */
};

