#include <stdlib.h>
#include <string.h>
#include <cdb.h>
#include <lx_string.h>
#include "misc.h"
#include "naw.h"
#include "naw_log.h"


int naw_data_init (data)
    struct naw_data *data;
{
    blobset *bs = blob_new(16, 0);
    if (!bs) return -1;

    data->bs = bs;
    return 0;
}

static void naw_data_free_elem (is)
    void *is;
{
    lx_free((lx_s *)is);
    free(is);
}

void naw_data_free (data)
    struct naw_data *data;
{
    blob_destroy(data->bs);
}

/***************************************************************************
  naw_data_cdb

  gather an entry from c and put it into the data hash

  returns 0 on success
  returns -1 if x could not be found in c
 ***************************************************************************/

static int naw_data_cdb (data, c, k, klen)
    struct naw_data *data;
    struct cdb *c;
    char *k;
    unsigned int klen;
{
    struct cdb_find cf;
    unsigned pos, len;
    int r = 0;
    ub1_t once=0;

    if (cdb_findinit(&cf, c, k, klen) < 0) {
        NAW_LOG(NAW_NOTICE, LXM_q("config: error initializing cdb"));
        return -1;
    }

    for (;;) {
        r = cdb_findnext(&cf);
        if (r == -1) {
            NAW_LOG(NAW_NOTICE, LXM_q("config: cdb short read"));
            return -1;
        } else if (r == 0) {
            if (once) break;
            NAW_LOG(NAW_NOTICE,
                    LXM_q("config: cdb data entry \""),
                    LXM_mb(k,klen),
                    LXM_q("\" does not exist"));
            return -1;
        }

        once = 1;

        pos = cdb_datapos(c);
        len = cdb_datalen(c);
        {
            char v[len];
        
            if (cdb_read(c, v, len, pos) == -1) {
                NAW_LOG(NAW_NOTICE,LXM_q("config: cdb short read"));
                return -1;
            }

            if (naw_data_put(data, v, len))
                return -1;
        }
    }

    return 0;
}


/***************************************************************************
  naw_data_parse

  parse a data entry into a stringlist
  environment entries are separated by commas
  variables include '=' within them, as a normal environment variable would,
  but needn't if they are to be used as booleans.

  returns 0 on success
  returns -1 on out-of-memory error
 ***************************************************************************/

int naw_data_parse (data, c, dp, dlen)
    struct naw_data *data;
    struct cdb *c;
    char *dp;
    unsigned int dlen;
{
    char *p, dp_len;

    NAW_LOG(NAW_DEBUG3, LXM_q("config: parsing data"));

    while (dlen) {
        p = memchr(dp, ',', dlen);
        if (p) {
            dp_len = p++ - dp;
            dlen  -= dp_len+1;
        } else {
            dp_len = dlen;
            dlen   = 0;
        }

        if (!dp_len) continue;
        
        if (*dp == '~') {
            if (naw_data_cdb(data, c, dp, dp_len))
                return -1;
        } else if (naw_data_put(data, dp, dp_len))
            return -1;
        dp = p;
    }

    NAW_LOG(NAW_DEBUG3, LXM_q("finished parsing data"));
    return 0;
}

/***************************************************************************
  naw_data_put

  sets the variable withing [data] to [p], which has length [len].
  [p] should be the full expression, "variable=value".

  Note that all variable/value entries will be null-terminated, but needn't
  be passed null-terminated.  This is for compatibility with C library calls
  only.

  returns  0 on success
  returns -1 on out-of-memory error.
 ***************************************************************************/

static inline int naw_data_setb(data, k, klen, v, vlen)
    struct naw_data *data;
    char *k, *v;
    unsigned int klen, vlen;
{
    lx_s *is;

    is = blob_get(data->bs, k, klen);

    if (!is) {
        is = calloc(1, sizeof(lx_s));
        if (!is) return -1;
        if (blob_register(data->bs, k, klen, is, naw_data_free_elem)) {
            free(is);
            return -1;
        }
    }

    if (lx_striset(is, v, vlen))
        return -1;

    if (lx_check0(is)) return -1;

    return 0;
}

int naw_data_set(data, k, v, vlen)
    struct naw_data *data;
    char *k, *v;
    unsigned int vlen;
{
    unsigned int klen = strlen(k);

    if (!v) {
        (void) blob_unregister(data->bs, k, klen);
        return 0;
    }

    if (!vlen) {
        v = "\0";
        vlen = 1;
    }

    return naw_data_setb(data, k, strlen(k), v, vlen);
}

int naw_data_put (data, k, len)
    struct naw_data *data;
    char *k;
    unsigned int len;
{
    unsigned int klen, vlen;
    char *v;

    v = memchr (k, '=', len);
    if (!v) {
        klen = len;
        v = "\0";
        vlen = 1;
    } else {
        klen = v++ - k;
        vlen = len - (klen + 1);
    }

    return naw_data_setb(data, k, klen, v, vlen);
}

/***************************************************************************
  naw_data_get

  searches the data structure referenced by [data] for a variable of name
  [s].  If the variable exists, *dstp and *dstlen are set to the memory
  location of the value and the length of the value, respectively.  Note
  that *dstp will be a direct pointer into memory used within [data].  If
  you need to play with this data, perhaps you'd like to duplicate it.

  Note also that because naw_data_put always null-terminates its data, *dstp
  will be null-terminated.  Note also that *dstlen will be the length of the
  data pre-null-termination, or technically the data length minus one.

  returns 0 if a variable was found; *dstp and *dstlen will be modified.
  returns 1 if no such variable was found; arguments will be left untouched.
 ***************************************************************************/

int naw_data_get (data, s, dstp, dstlen)
    struct naw_data *data;
    char *s;
    char **dstp;
    unsigned int *dstlen;
{

    lx_s *is;

    is = blob_get(data->bs, s, strlen(s));
    if (!is) return 1;

    *dstp   = is->s;
    *dstlen = is->len-1;
    return 0;
}
